//
// expr.h
//
#ifndef __EXPR__
#define __EXPR__

#include "refobj.h"
#include "varstore.h"

typedef struct expr {
    refobj_t ref;   //ref is at the beginning of rat
    void   ( *eval  )(struct expr *self, var_store_t *store);
    void   ( *print )(struct expr *self);
} expr_t;

typedef struct name_list_entry {
    char *name;
    list_t lst;
} name_list_entry_t;

typedef struct expr_list_entry {
    expr_t *expr;
    list_t lst;
} expr_list_entry_t;

//number
extern expr_t *expr_make_num(int n);

//variable
extern expr_t *expr_make_var(char *id);

//arithmetic operators
extern expr_t *expr_make_add(expr_t *a, expr_t *b);
extern expr_t *expr_make_sub(expr_t *a, expr_t *b);
extern expr_t *expr_make_mul(expr_t *a, expr_t *b);
extern expr_t *expr_make_div(expr_t *a, expr_t *b);

//comparators
extern expr_t *expr_make_eq(expr_t *a, expr_t *b);
extern expr_t *expr_make_ne(expr_t *a, expr_t *b);
extern expr_t *expr_make_ge(expr_t *a, expr_t *b);
extern expr_t *expr_make_gt(expr_t *a, expr_t *b);
extern expr_t *expr_make_le(expr_t *a, expr_t *b);
extern expr_t *expr_make_lt(expr_t *a, expr_t *b);

//function application
extern expr_t *expr_make_fun_app(char *fname, list_t *args);
#endif
