//
// expr_fun_app.c
//  function application
//
#include "common.h"
#include "func.h"
#include "varstore.h"
#include <string.h>

typedef struct expr_func_app {
    refobj_t ref;   //ref is at the beginning of stmt
    void  ( *eval  )(struct expr *self, var_store_t *store);
    void ( *print )(struct expr *self);

    char *fname;    //function name
    list_t *args;   //actual parameters of expr_list_entry_t list
} expr_func_app_t;

static void eval_func_app(expr_t *self, var_store_t *store) {
    ON_FALSE_EXIT(self->ref.tag == OBJ_EXPR_FUNC_APP,
                  strmsg("tag (%d) is not OBJ_EXPR_FUNC_APP", self->ref.tag));
    expr_func_app_t *expr = (expr_func_app_t*) self;

    //evaluate the actual parameters in the reverse order and push the results to the stack
    for(list_t *pos = expr->args->prev ; pos != expr->args; pos = pos->prev) {
        expr_list_entry_t *arg = containerof(pos, expr_list_entry_t, lst);
        //TODO: evaluate the actual parameter
        //hint: use arg->expr->eval, this call will push the result onto the stack

    }
    
    COMMENT("", expr->print(self), "");
    //TODO: invoke the function
    //hint: use expr->fname


    //TODO: remove the actual parameters from the stack
    //hint: use list_size(expr->args)


    //TODO: push the result to the stack
    //hint: the resut is in rax

}

static void print_func_app(expr_t *self) {
    ON_FALSE_EXIT(self->ref.tag == OBJ_EXPR_FUNC_APP,
                  strmsg("tag (%d) is not OBJ_EXPR_FUNC_APP", self->ref.tag));
    expr_func_app_t *expr = (expr_func_app_t*) self;

    printf("%s(", expr->fname);
    for(list_t *pos = expr->args->next; pos != expr->args; pos = pos->next) {
        expr_t *arg = containerof(pos, expr_list_entry_t, lst)->expr;
        arg->print(arg);
        printf("%s", pos->next != expr->args? ", ": "");
    }
    printf(")");
}

static void release_func_app(refobj_t *ref) {
    ON_FALSE_EXIT(ref->tag == OBJ_EXPR_FUNC_APP,
                  strmsg("tag (%d) is not OBJ_EXPR_FUNC_APP", ref->tag));    
    expr_func_app_t *expr = (expr_func_app_t*) ref;

    refobj_decref(&expr->ref);
    if(expr->ref.cnt_ref == 0) {
        free(expr->fname);
        for(list_t *pos = expr->args->next; pos != expr->args; ) {
            expr_list_entry_t *entry = containerof(pos, expr_list_entry_t, lst);
            pos = pos->next;
            entry->expr->ref.release(&entry->expr->ref);
            free(entry);
        }        
        free(expr->args);
        refobj_free(&expr->ref);
    }
}

expr_t *expr_make_fun_app(char *fname, list_t *args) {
    expr_func_app_t* expr = refobj_alloc(OBJ_EXPR_FUNC_APP, sizeof(expr_func_app_t));
    expr->ref.release = release_func_app;
    expr->eval        = eval_func_app;
    expr->print       = print_func_app;

    expr->fname = strdup(fname);
    expr->args = args;
    
    for(list_t *pos = expr->args->next; pos != expr->args; pos = pos->next) {
        expr_list_entry_t *entry = containerof(pos, expr_list_entry_t, lst);
        entry->expr->ref.addref(&entry->expr->ref);
    }        

    return (expr_t*)expr;
}
