//
// expr_var.c
//  variables
//

#include "expr.h"
#include "varstore.h"
#include <string.h>

typedef struct expr_var {
    refobj_t ref;   //first part is the same as expr_t
    void   ( *eval  )(struct expr *self, var_store_t *store);
    void   ( *print )(struct expr *self);

    char *id;       //variable
} expr_var_t;

static void eval_var(expr_t *self, var_store_t *store) {
    ON_FALSE_EXIT(self->ref.tag == OBJ_EXPR_VAR,
                  strmsg("tag (%d) is not OBJ_EXPR_VAR", self->ref.tag));

    expr_var_t *expr = (expr_var_t*) self;

    COMMENT(strmsg("%s", expr->id), , "");
    if(!var_store_has(store, expr->id))
        store = NULL;   //global variable
    int inx = var_store_get(store, expr->id); //get the index of the variable
    
    if(store == NULL) {
        //TODO: handle as a global variable
        //inx is a negative number. making it positive,
        //variable address is VAR - inx*8 from rip

    }
    else if(inx > 0) {  
        //TODO: handle as a parameter 
        //accounting for 16 bytes for return address and rbp and
        //accounting for 8 bytes for the push op (memory at rsp has data),
        //variable address is 16 + inx*8 - 8 from rbp

    }
    else {
        //TODO: handle as a local variable 
        //accounting for 8 bytes for the push op (memory at rsp has data),
        //variable address is inx*8 - 8 from rbp

    }
}

static void print_var(expr_t *self) {
    ON_FALSE_EXIT(self->ref.tag == OBJ_EXPR_VAR,
                  strmsg("tag (%d) is not OBJ_EXPR_VAR", self->ref.tag));
                  
    expr_var_t *expr = (expr_var_t*) self;
    printf("%s", expr->id);
}

static void release_var(refobj_t *ref) {
    ON_FALSE_EXIT(ref->tag == OBJ_EXPR_VAR,
                  strmsg("tag (%d) is not OBJ_EXPR_VAR", ref->tag));    

    expr_var_t *expr = (expr_var_t*) ref;
    refobj_decref(&expr->ref);
    if(expr->ref.cnt_ref == 0) {
        free(expr->id);
        refobj_free(&expr->ref);
    }
}

expr_t *expr_make_var(char *id) {
    expr_var_t* expr  = refobj_alloc(OBJ_EXPR_VAR, sizeof(expr_var_t));
    expr->ref.release = release_var;
    expr->eval        = eval_var;
    expr->print       = print_var;

    expr->id  = strdup(id);
    return (expr_t*)expr;
}

