#include "func.h"
#include <stdio.h>
#include <string.h>

static void exec_func(func_t *self) {
    ON_FALSE_EXIT(self->ref.tag == OBJ_FUNC,
                  strmsg("tag (%d) is not OBJ_FUNC", self->ref.tag));
    func_t *func = self;

    //create a local_store for this function
    var_store_t store;
    var_store_init(&store, list_size(func->ids));

    //add parameters to store
    for(list_t *pos = func->ids->prev; pos != func->ids; pos = pos->prev) {
        name_list_entry_t *entry = containerof(pos, name_list_entry_t, lst);
        var_store_set(&store, entry->name);
    }

    //add local variables to store
    func->body->addvar(func->body, &store);

    COMMENT(strmsg("fun %s(", func->fname), {
                for(list_t *pos = func->ids->next; pos != func->ids; pos = pos->next) {
                    name_list_entry_t *entry = containerof(pos, name_list_entry_t, lst);
                    printf("%s", entry->name);
                    printf("%s", pos->next != func->ids? ", ": "");
                }        
            }, ")");

    //TODO: generate the preamble for the function
    //   1. add a label with the function name (hint: use func->fname)
    //   2. save the caller's stack frame pointer
    //   3. set the callee's stack frame pointer
    //   4. allocate stack space for local variables (hint: var_store_local_var_size)


    //TODO: generate the body of the function
    //hint: use func->body->exec


    COMMENT(strmsg("exit %s", func->fname), , "");
    //TODO: generate the postamble for the function
    //hint: leave then return


    //destroy the local_store for this function
    var_store_destroy(&store);
}

static void print_func(func_t *self) {
    ON_FALSE_EXIT(self->ref.tag == OBJ_FUNC,
                  strmsg("tag (%d) is not OBJ_FUNC", self->ref.tag));
    func_t *func = self;
    printf("%s(", func->fname);
    for(list_t *pos = func->ids->next; pos != func->ids; pos = pos->next) {
        name_list_entry_t *entry = containerof(pos, name_list_entry_t, lst);
        printf("%s", entry->name);
        printf("%s", pos->next != func->ids? ", ": "");
    }
    printf(")\n");
    func->body->print(func->body);
}

static void release_func(refobj_t *ref) {
    ON_FALSE_EXIT(ref->tag == OBJ_FUNC,
                  strmsg("tag (%d) is not OBJ_FUNC", ref->tag));
                  
    func_t *func = (func_t*) ref;
    refobj_decref(&func->ref);
    if(func->ref.cnt_ref == 0) {
        free(func->fname);
        for(list_t *pos = func->ids; pos != func->ids; ) {
            name_list_entry_t *entry = containerof(pos, name_list_entry_t, lst);
            pos = pos->next;
            free(entry->name);
            free(entry);
        }
        free(func->ids);
        func->body->ref.release(&func->body->ref);
        refobj_free(&func->ref);
    }
}

extern func_t *func_make(char *fname, list_t *ids, stmt_t *body) {
    func_t* func = refobj_alloc(OBJ_FUNC, sizeof(func_t));
    func->ref.release = release_func;
    func->exec  = exec_func;
    func->print = print_func;

    func->fname = strdup(fname);
    func->ids = ids;
    func->body = body;
    body->ref.addref(&body->ref);
    return func;
}
