//
// program.c
//  represents the entire program
//

#include "common.h"
#include "refobj.h"
#include "program.h"
#include "varstore.h"
#include <stdio.h>

static void header() {
    printf("    .global main\n");
    printf("    .text\n");
}

static void footer() {
    extern void read_func();
    read_func();
    
    extern void write_func();
    write_func();

    printf("    .section .rodata\n");
    printf("ANSW:       .string \"answer: %%ld\\n\"\n");
    printf("ENTR:       .string \"enter: \"\n");
    printf("ENTR_FMT:   .string \"%%ld\"\n");

    if( var_store_local_var_size(NULL) > 0) {
        //TODO: allocate global variables
        //hint: these global variables are in .data section
        //hint: use VAR label for the variable name
        //hint: use .zero x to allocate x bytes
        //hint: use var_store_local_var_size(NULL)


    }
}

static void exec_program(program_t *self) {
    ON_FALSE_EXIT(self->ref.tag == OBJ_PROGRAM,
                  strmsg("tag (%d) is not OBJ_PROGRAM", self->ref.tag));

    program_t *pgm = self;

    //init the global variable store
    var_store_init(NULL /*global*/, 0);

    //add global variables to the global store
    for(list_t *pos = pgm->vars->next; pos != pgm->vars; pos = pos->next) {
        name_list_entry_t *entry = containerof(pos, name_list_entry_t, lst);
        var_store_set(NULL, entry->name);
    }

    //TODO: generate header


    //translate function definitions
    for(list_t *pos = pgm->funcs->next; pos != pgm->funcs; pos = pos->next) {
        func_list_entry_t *entry = containerof(pos, func_list_entry_t, lst);
        //TODO: generate code for the function entry->func
        //hint: use entry->func->exec


    }

    //TODO: generate footer


    //destroy the global variable store
    var_store_destroy(NULL /*global*/);
}

static void print_program(program_t *self) {
    ON_FALSE_EXIT(self->ref.tag == OBJ_PROGRAM,
                  strmsg("tag (%d) is not OBJ_PROGRAM", self->ref.tag));
    program_t *pgm = self;

    printf("vars ");
    for(list_t *pos = pgm->vars->next; pos != pgm->vars; pos = pos->next) {
        name_list_entry_t *entry = containerof(pos, name_list_entry_t, lst);
        printf("%s", entry->name);
        printf("%s", pos->next != pgm->vars? ", ": "\n");
    }

    for(list_t *pos = pgm->funcs->next; pos != pgm->funcs; pos = pos->next) {
        func_list_entry_t *entry = containerof(pos, func_list_entry_t, lst);
        entry->func->print(entry->func);
    }
}

static void release_program(refobj_t *ref) {
    ON_FALSE_EXIT(ref->tag == OBJ_PROGRAM,
                  strmsg("tag (%d) is not OBJ_PROGRAM", ref->tag));
    program_t *pgm = (program_t*)ref;
    
    refobj_decref(&pgm->ref);
    if(pgm->ref.cnt_ref == 0) {
        //release global variables
        for(list_t *pos = pgm->vars; pos != pgm->vars; ) {
            name_list_entry_t *entry = containerof(pos, name_list_entry_t, lst);
            pos = pos->next;
            free(entry->name);
            free(entry);
        }
        free(pgm->vars);

        //release func defintions
        for(list_t *pos = pgm->funcs->next; pos != pgm->funcs; ) {
            func_list_entry_t *entry = containerof(pos, func_list_entry_t, lst);
            pos = pos->next;
            entry->func->ref.release(&entry->func->ref);
            free(entry);
        }
        free(pgm->funcs);
        refobj_free(&pgm->ref);
    }
}

program_t *program_make(list_t *vars, list_t *funcs) {
    program_t *pgm = refobj_alloc(OBJ_PROGRAM, sizeof(program_t));
    pgm->ref.release  = release_program;
    pgm->exec         = exec_program;
    pgm->print        = print_program;

    pgm->vars  = vars;
    pgm->funcs = funcs;
    for(list_t *pos = pgm->funcs->next; pos != pgm->funcs; pos = pos->next) {
        func_list_entry_t *entry = containerof(pos, func_list_entry_t, lst);
        entry->func->ref.addref(&entry->func->ref);
    }
    return pgm;
}
