//
// stmt_assignment.c
//  assignment statement
//

#include "common.h"
#include "refobj.h"
#include "expr.h"
#include "stmt.h"
#include "varstore.h"
#include <stdio.h>
#include <string.h>

typedef struct stmt_assignment {
    refobj_t ref;   //ref is at the beginning of stmt
    void   ( *exec   )(struct stmt *self, var_store_t *store);
    void   ( *addvar )(struct stmt *self, var_store_t *store);
    void   ( *print  )(struct stmt *self);

    char *id;
    expr_t *rhs;
} stmt_assignment_t;

static void exec_assignment(stmt_t *self, var_store_t *store) {
    ON_FALSE_EXIT(self->ref.tag == OBJ_STMT_ASSIGN,
                  strmsg("tag (%d) is not OBJ_STMT_ASSIGN", self->ref.tag));

    stmt_assignment_t *stmt = (stmt_assignment_t*) self;

    //eval rhs
    //TODO: evaluate stmt->rhs
    //hint: use stmt->rhs->eval


    //assignment
    COMMENT(strmsg("%s := ", stmt->id), stmt->rhs->print(stmt->rhs), "");
    if(!var_store_has(store, stmt->id))
        store = NULL;   //global variable
    int inx = var_store_get(store, stmt->id); //get the index of the variable

    //copy the result of rhs to the variable
    if(store == NULL) {
        //TODO: handle as a global variable
        //inx is a negative number. making it positive,
        //variable address is VAR - inx*8 from rip

    }
    else if(inx > 0) {  
        //TODO: handle as a parameter 
        //accounting for 16 bytes for return address and rbp and
        //accounting for 8 bytes for the push op (memory at rsp has data),
        //variable address is 16 + inx*8 - 8 from rbp

    }
    else {
        //TODO: handle as a local variable 
        //accounting for 8 bytes for the push op (memory at rsp has data),
        //variable address is inx*8 - 8 from rbp

    }
}

static void addvar_assignment(stmt_t *self, var_store_t *store) {
    ON_FALSE_EXIT(self->ref.tag == OBJ_STMT_ASSIGN,
                  strmsg("tag (%d) is not OBJ_STMT_ASSIGN", self->ref.tag));

    stmt_assignment_t *stmt = (stmt_assignment_t*) self;
    if(!var_store_has(store, stmt->id)) {
        if(!var_store_has(NULL, stmt->id))
            var_store_set(store, stmt->id); //new local variable
    }
}

static void print_assignment(stmt_t *self) {
    ON_FALSE_EXIT(self->ref.tag == OBJ_STMT_ASSIGN,
                  strmsg("tag (%d) is not OBJ_STMT_ASSIGN", self->ref.tag));

    stmt_assignment_t *stmt = (stmt_assignment_t*) self;
    printf("%s := ", stmt->id);
    stmt->rhs->print(stmt->rhs);
    printf("\n");
}

static void release_assignment(refobj_t *ref) {
    ON_FALSE_EXIT(ref->tag == OBJ_STMT_ASSIGN,
                  strmsg("tag (%d) is not OBJ_STMT_ASSIGN", ref->tag));
                  
    stmt_assignment_t *stmt = (stmt_assignment_t*) ref;
    refobj_decref(&stmt->ref);
    if(stmt->ref.cnt_ref == 0) {
        free(stmt->id);
        stmt->rhs->ref.release(&stmt->rhs->ref);
        refobj_free(&stmt->ref);
    }
}

stmt_t *stmt_make_assignment(char *id, expr_t *rhs) {
    stmt_assignment_t* stmt = refobj_alloc(OBJ_STMT_ASSIGN, sizeof(stmt_assignment_t));
    stmt->ref.release       = release_assignment;
    stmt->exec              = exec_assignment;
    stmt->addvar            = addvar_assignment;
    stmt->print             = print_assignment;

    stmt->id  = strdup(id);
    stmt->rhs = rhs;
    rhs->ref.addref(&rhs->ref);
    return (stmt_t*) stmt;
}
