//
// stmt_read.c
//  read statement
//

#include "common.h"
#include "refobj.h"
#include "expr.h"
#include "stmt.h"
#include "varstore.h"
#include <stdio.h>
#include <string.h>

typedef struct stmt_read {
    refobj_t ref;   //ref is at the beginning of stmt
    void   ( *exec   )(struct stmt *self, var_store_t *store);
    void   ( *addvar )(struct stmt *self, var_store_t *store);
    void   ( *print  )(struct stmt *self);

    char *id;
} stmt_read_t;


static void exec_read(stmt_t *self, var_store_t *store) {
    ON_FALSE_EXIT(self->ref.tag == OBJ_STMT_READ,
                  strmsg("tag (%d) is not OBJ_STMT_READ", self->ref.tag));

    stmt_read_t *stmt = (stmt_read_t*) self;

    COMMENT(strmsg("read %s", stmt->id), , "");

    if(!var_store_has(store, stmt->id))
        store = NULL;   //global variable
    int inx = var_store_get(store, stmt->id); //get the index of the variable

    //load the ADDRESS of the variable to %rax
    if(store == NULL) {
        //TODO: handle as a global variable
        //inx is a negative number. making it positive,
        //variable address is VAR - inx*8 from rip

    }
    else if(inx > 0) {  
        //TODO: handle as a parameter 
        //accounting for 16 bytes for return address and rbp and
        //accounting for 8 bytes for the push op (memory at rsp has data),
        //variable address is 16 + inx*8 - 8 from rbp

    }
    else {
        //TODO: handle as a local variable 
        //accounting for 8 bytes for the push op (memory at rsp has data),
        //variable address is inx*8 - 8 from rbp

    }
    
    printf("    pushq   %%rax\n");
    printf("    call    read_fun\n");
    printf("    addq    $8, %%rsp\n");
}

static void addvar_read(stmt_t *self, var_store_t *store) {
    ON_FALSE_EXIT(self->ref.tag == OBJ_STMT_READ,
                  strmsg("tag (%d) is not OBJ_STMT_READ", self->ref.tag));

    stmt_read_t *stmt = (stmt_read_t*) self;
    if(!var_store_has(store, stmt->id)) {
        if(!var_store_has(NULL, stmt->id))
            var_store_set(store, stmt->id); //new local variable
    }
}

static void print_read(stmt_t *self) {
    ON_FALSE_EXIT(self->ref.tag == OBJ_STMT_READ,
                  strmsg("tag (%d) is not OBJ_STMT_READ", self->ref.tag));

    stmt_read_t *stmt = (stmt_read_t*) self;
    printf("read %s\n", stmt->id);
}

static void release_read(refobj_t *ref) {
    ON_FALSE_EXIT(ref->tag == OBJ_STMT_READ,
                  strmsg("tag (%d) is not OBJ_STMT_READ", ref->tag));

    stmt_read_t *stmt = (stmt_read_t*) ref;
    refobj_decref(&stmt->ref);
    if(stmt->ref.cnt_ref == 0) {
        free(stmt->id);
        refobj_free(&stmt->ref);
    }
}

stmt_t *stmt_make_read(char *id) {
    stmt_read_t* stmt = refobj_alloc(OBJ_STMT_READ, sizeof(stmt_read_t));
    stmt->ref.release = release_read;
    stmt->exec        = exec_read;
    stmt->addvar      = addvar_read;
    stmt->print       = print_read;

    stmt->id = strdup(id);
    return (stmt_t*) stmt;
}

void read_func() {
    printf("read_fun:\n");
    printf("    pushq   %%rbp\n");    
    printf("    movq    %%rsp, %%rbp\n");    
    
    //AMD64 System V ABI: make rsp a multiple of 16
    printf("    andq    $~15, %%rsp\n");    

    //print the enter: message
    printf("    leaq    ENTR(%%rip), %%rdi\n");
    printf("    movl    $0, %%eax\n");
    printf("    call    printf\n");
    
    //read a number
    printf("    movq    16(%%rbp), %%rsi\n");    
    printf("    leaq    ENTR_FMT(%%rip), %%rdi\n");
    printf("    movl    $0, %%eax\n");
    printf("    call    scanf\n");

    //return
    printf("    leave\n");
    printf("    ret\n");
}

