//
// stmt_return.c
//  return statement
//

#include "stmt.h"
#include "varstore.h"
#include <stdio.h>

typedef struct stmt_return {
    refobj_t ref;   //ref is at the beginning of stmt
    void   ( *exec   )(struct stmt *self, var_store_t *store);
    void   ( *addvar )(struct stmt *self, var_store_t *store);
    void   ( *print  )(struct stmt *self);

    expr_t *expr;
} stmt_return_t;


static void exec_return(stmt_t *self, var_store_t *store) {
    ON_FALSE_EXIT(self->ref.tag == OBJ_STMT_RETURN,
                  strmsg("tag (%d) is not OBJ_STMT_RETURN", self->ref.tag));

    stmt_return_t *stmt = (stmt_return_t*) self;

    //TODO: evalate stmt->expr 
    //hint: use stmt->expr->eval   


    COMMENT("return ", stmt->expr->print(stmt->expr), "");

    //TODO: pop the result of stmt->expr to rax
    //hint: the result is in the stack


    //TODO: deallocate the stack space for local variables
    //hint: rbp has the stack pointer value before the allocation


    //TODO: restore the caller's stack frame pointer


    //TODO: return from this function

}

static void addvar_return(stmt_t *self, var_store_t *store) {
    ON_FALSE_EXIT(self->ref.tag == OBJ_STMT_RETURN,
                  strmsg("tag (%d) is not OBJ_STMT_RETURN", self->ref.tag));
}

static void print_return(stmt_t *self) {
    ON_FALSE_EXIT(self->ref.tag == OBJ_STMT_RETURN,
                  strmsg("tag (%d) is not OBJ_STMT_RETURN", self->ref.tag));

    stmt_return_t *stmt = (stmt_return_t*) self;
    printf("return ");
    stmt->expr->print(stmt->expr);
    printf("\n");
}

static void release_return(refobj_t *ref) {
    ON_FALSE_EXIT(ref->tag == OBJ_STMT_RETURN,
                  strmsg("tag (%d) is not OBJ_STMT_RETURN", ref->tag));
                  
    stmt_return_t *stmt = (stmt_return_t*) ref;
    refobj_decref(&stmt->ref);
    if(stmt->ref.cnt_ref == 0) {
        stmt->expr->ref.release(&stmt->expr->ref);
        refobj_free(&stmt->ref);
    }
}

stmt_t *stmt_make_return(expr_t *expr) {
    stmt_return_t* stmt = refobj_alloc(OBJ_STMT_RETURN, sizeof(stmt_return_t));
    stmt->ref.release   = release_return;
    stmt->exec          = exec_return;
    stmt->addvar        = addvar_return;
    stmt->print         = print_return;

    stmt->expr = expr;
    expr->ref.addref(&expr->ref);
    return (stmt_t*) stmt;
}
