//
// stmt_while.c
//  while statement
//

#include "common.h"
#include "refobj.h"
#include "expr.h"
#include "stmt.h"
#include <stdio.h>

typedef struct stmt_while {
    refobj_t ref;   //ref is at the beginning of stmt
    void   ( *exec   )(struct stmt *self, var_store_t *store);
    void   ( *addvar )(struct stmt *self, var_store_t *store);
    void   ( *print  )(struct stmt *self);

    expr_t *cond;
    stmt_t *loop_stmt;
} stmt_while_t;

static void exec_while(stmt_t *self, var_store_t *store) {
    ON_FALSE_EXIT(self->ref.tag == OBJ_STMT_WHILE,
                  strmsg("tag (%d) is not OBJ_STMT_WHILE", self->ref.tag));

    stmt_while_t *stmt = (stmt_while_t*) self;

    int label_test = label_new();
    int label_exit = label_new();

    //evaluate condition
    COMMENT("while ( ", stmt->cond->print(stmt->cond), " ) ...");
    //TODO: print label_test and evaluate stmt->cond
    //hint: use label_str function and stmt->cond->eval


    //jump to label_exit if false
    COMMENT("while ( ", stmt->cond->print(stmt->cond), " )");
    //TODO: jump to label_exit if the evaluation result is 0
    //hint: use label_str function


    //execute the body
    //TODO: execute stmt->loop_stmt
    //hint: use stmt->loop_stmt->exec


    //end the while loop
    //TODO: jump to label_test
    //hint: use label_str function


    COMMENT("end while ( ", stmt->cond->print(stmt->cond), " ) ...");
    //TODO: print the label_exit
    //hint: use label_str function

}

static void addvar_while(stmt_t *self, var_store_t *store) {
    ON_FALSE_EXIT(self->ref.tag == OBJ_STMT_WHILE,
                  strmsg("tag (%d) is not OBJ_STMT_WHILE", self->ref.tag));

    stmt_while_t *stmt = (stmt_while_t*) self;
    stmt->loop_stmt->addvar(stmt->loop_stmt, store);
}

static void print_while(stmt_t *self) {
    ON_FALSE_EXIT(self->ref.tag == OBJ_STMT_WHILE,
                  strmsg("tag (%d) is not OBJ_STMT_WHILE", self->ref.tag));

    stmt_while_t *stmt = (stmt_while_t*) self;
    printf("while ( ");
    stmt->cond->print(stmt->cond);
    printf(" )\n");
    stmt->loop_stmt->print(stmt->loop_stmt);
}

static void release_while(refobj_t *ref) {
    ON_FALSE_EXIT(ref->tag == OBJ_STMT_WHILE,
                  strmsg("tag (%d) is not OBJ_STMT_WHILE", ref->tag));
                  
    stmt_while_t *stmt = (stmt_while_t*) ref;
    refobj_decref(&stmt->ref);
    if(stmt->ref.cnt_ref == 0) {
        stmt->cond->ref.release(&stmt->cond->ref);
        stmt->loop_stmt->ref.release(&stmt->loop_stmt->ref);
        refobj_free(&stmt->ref);
    }
}

stmt_t *stmt_make_while(expr_t *cond, stmt_t *loop_stmt) {
    stmt_while_t* stmt = refobj_alloc(OBJ_STMT_WHILE, sizeof(stmt_while_t));
    stmt->ref.release  = release_while;
    stmt->exec         = exec_while;
    stmt->addvar       = addvar_while;
    stmt->print        = print_while;

    stmt->cond = cond;
    stmt->loop_stmt = loop_stmt;
    cond->ref.addref(&cond->ref);
    loop_stmt->ref.addref(&loop_stmt->ref);
    return (stmt_t*) stmt;
}
