//
// stmt_write.c
//  write statement
//

#include "common.h"
#include "refobj.h"
#include "expr.h"
#include "stmt.h"
#include <stdio.h>

typedef struct stmt_write {
    refobj_t ref;   //ref is at the beginning of stmt
    void   ( *exec   )(struct stmt *self, var_store_t *store);
    void   ( *addvar )(struct stmt *self, var_store_t *store);
    void   ( *print  )(struct stmt *self);

    expr_t *expr;
} stmt_write_t;


static void exec_write(stmt_t *self, var_store_t *store) {
    ON_FALSE_EXIT(self->ref.tag == OBJ_STMT_WRITE,
                  strmsg("tag (%d) is not OBJ_STMT_WRITE", self->ref.tag));

    stmt_write_t *stmt = (stmt_write_t*) self;

    //evaluate expression
    stmt->expr->eval(stmt->expr, store);

    //print the result
    COMMENT("write ", stmt->expr->print(stmt->expr), "");
    //skip popq rax, pushq rax
    printf("    call    write_fun\n");
    printf("    addq    $8, %%rsp\n");    
}

static void addvar_write(stmt_t *self, var_store_t *store) {
    ON_FALSE_EXIT(self->ref.tag == OBJ_STMT_WRITE,
                  strmsg("tag (%d) is not OBJ_STMT_WRITE", self->ref.tag));
}

static void print_write(stmt_t *self) {
    ON_FALSE_EXIT(self->ref.tag == OBJ_STMT_WRITE,
                  strmsg("tag (%d) is not OBJ_STMT_WRITE", self->ref.tag));

    stmt_write_t *stmt = (stmt_write_t*) self;
    printf("write ");
    stmt->expr->print(stmt->expr);
    printf("\n");
}

static void release_write(refobj_t *ref) {
    ON_FALSE_EXIT(ref->tag == OBJ_STMT_WRITE,
                  strmsg("tag (%d) is not OBJ_STMT_WRITE", ref->tag));
                  
    stmt_write_t *stmt = (stmt_write_t*) ref;
    refobj_decref(&stmt->ref);
    if(stmt->ref.cnt_ref == 0) {
        stmt->expr->ref.release(&stmt->expr->ref);
        refobj_free(&stmt->ref);
    }
}

stmt_t *stmt_make_write(expr_t *expr) {
    stmt_write_t* stmt = refobj_alloc(OBJ_STMT_WRITE, sizeof(stmt_write_t));
    stmt->ref.release  = release_write;
    stmt->exec         = exec_write;
    stmt->addvar       = addvar_write;
    stmt->print        = print_write;

    stmt->expr = expr;
    expr->ref.addref(&expr->ref);
    return (stmt_t*) stmt;
}

void write_func() {
    printf("write_fun:\n");
    printf("    pushq   %%rbp\n");    
    printf("    movq    %%rsp, %%rbp\n");    

    //AMD64 System V ABI: make rsp a multiple of 16
    printf("    andq    $~15, %%rsp\n");    
    
    //write a number
    printf("    movq    16(%%rbp), %%rsi\n");    
    printf("    leaq    ANSW(%%rip), %%rdi\n");
    printf("    movl    $0, %%eax\n");
    printf("    call    printf\n");

    //return
    printf("    leave\n");
    printf("    ret\n");
}
