//
// varstore.c
//  variable store
//
#include "common.h"
#include "list.h"
#include "varstore.h"
#include <stdio.h>
#include <string.h>

typedef struct var_entry {
    char *name;
    int  value;

    list_t lst;
} var_entry_t;

//forward definitions
static void var_store_add(var_store_t *store, char *name, int value);
static var_entry_t *var_store_find(var_store_t *store, char *name);
static var_entry_t *var_store_find_from_store(var_store_t *store, char *name);

static var_store_t store_global;

void var_store_init(var_store_t *store, int cnt_param) {
    if(store == NULL)
        store = &store_global;
    store->cnt_param = cnt_param;
    list_init_head(&store->lst);
}

void var_store_destroy(var_store_t *store) {
    if(store == NULL)
        store = &store_global;
    for(list_t *pos = store->lst.next; pos != &store->lst; ) {
        var_entry_t *var = containerof(pos, var_entry_t, lst);
        pos = pos->next;
        free(var);
    }
}

int var_store_has(var_store_t *store, char *name) {
    if(store == NULL)
        store = &store_global;
    var_entry_t *var = var_store_find(store, name);
    return var != NULL;
}

int var_store_get(var_store_t *store, char *name) {
    if(store == NULL)
        store = &store_global;
    var_entry_t *var = var_store_find(store, name);
    ON_FALSE_EXIT(var != NULL, strmsg("cannot find variable %s", name));
    return var->value;
}

int var_store_set(var_store_t *store, char *name) {
    if(store == NULL)
        store = &store_global;
    var_entry_t *var = var_store_find(store, name);
    if(var != NULL)
        return var->value;
    else {
        int index = store->cnt_param - list_size(&store->lst);
        var_store_add(store, name, index);
        return index;
    }
}

int  var_store_parameter_size(var_store_t *store) {
    if(store == NULL)
        store = &store_global;
    return store->cnt_param;
}

int  var_store_local_var_size(var_store_t *store) {
    if(store == NULL)
        store = &store_global;
    return list_size(&store->lst) - store->cnt_param;
}

static void var_store_add(var_store_t *store, char *name, int value) {
    if(store == NULL)
        store = &store_global;
    var_entry_t *var = malloc(sizeof(var_entry_t));
    var->name = name;
    var->value = value;
    list_add_to_last(&store->lst, &var->lst);
}

static var_entry_t *var_store_find(var_store_t *store, char *name) {
    list_t *head = &store->lst;
    list_t *pos;
    for(pos = head->next; pos != head; pos = pos->next) {
        var_entry_t *var = containerof(pos, var_entry_t, lst);
        if(strcmp(var->name, name) == 0)
            return var;
    }
    return NULL;   
}
